#include <iostream> // cin, cout
#include <cmath> // fabs
#include <memory> // memcpy
#include <GL/glut.h>

using namespace std;

///////////////////////////////////////////////////////////
// clase: puntos de cuatro floats
// 4 para poder hacer racionales y porque GLU pide 3D o 4D
// floats porque GLU pide floats
// tiene solo lo necesario para usarlos aca
// z siempre es 0
const size_t SZF=sizeof(float);
class p4f{
  float c[4];
public:
  p4f(double x=0, double y=0, double z=0, double w=1){
    c[0]=x; c[1]=y; c[2]=z; c[3]=w;
  }
  p4f(const p4f &p){memcpy(c,p.c,4*SZF);}

  operator const float*() const {return c;}

  p4f& operator=(const p4f &p){memcpy(c,p.c,4*SZF); return *this;}

  // modifica el peso pero mantiene el punto en el mismo lugar de R3
  // ==> OJO: no admite peso 0 !!!!!!!!!!!!!!!!!
  p4f& peso(float w){
    if (fabs(w)<1e-7) {if (w<0) w=-1e-7; else w=1e-7;}
    c[0]*=w/c[3]; c[1]*=w/c[3]; c[2]*=w/c[3]; c[3]=w; return *this;
  }

  bool cerca2D(const p4f &p, float r=1){
    float cw=c[3],pw=p[3];
    return (fabs(c[0]/cw-p[0]/pw)<=r && fabs(c[1]/cw-p[1]/pw)<=r);
  }

  float& operator[](int i) {return c[i];}
  const float& operator[](int i) const {return c[i];}

  p4f& neg(){c[0]=-c[0]; c[1]=-c[1]; c[2]=-c[2]; c[3]=-c[3]; return *this;}
  p4f operator -() const {return p4f(-c[0],-c[1],-c[2],-c[3]);}

  // asigna el punto interpolado
  p4f& lerp(const p4f& p0,const p4f& p1,float u){
    c[0]=(1-u)*p0[0]+u*p1[0];
    c[1]=(1-u)*p0[1]+u*p1[1];
    c[2]=(1-u)*p0[2]+u*p1[2];
    c[3]=(1-u)*p0[3]+u*p1[3];
    return *this;
  }
};
///////////////////////////////////////////////////////////

// variables
int 
  w,h, // alto y ancho de la pantalla
  tol=10, // tolerancia para seleccionar un punto
  yclick, // y picado
  lod=32; // nivel de detalle (subdivisiones de lineas)

float // colores
  fondo[]={0.95f,0.98f,1.0f}, // color de fondo
  lc_c[]={.4f,.4f,.6f},       // poligono de control
  pc_c[]={.2f,.2f,.4f},       // puntos de control
  c_c[]={.8f,.4f,.4f};        // curva

bool
  mini=false,      // si esta minimizado
  cl_info=true,    // informa por la linea de comandos;
  antialias=true;  // antialiasing

// NURBS
int MAXPC=256, MAXORDER=0;
int uo=3; // orden (grado+1) en u
p4f *pc=0; int npc=0; // puntos de control
float* ukv; int unk=0; // vector y cant de knots en u
bool periodico=false, BezierBC=true; // periodico y condiciones de borde Bezier
GLUnurbsObj *curva=0; // se crea al inicializar y se destruye al salir
int pcsel=-1; // punto seleccionado
float W0; // peso

// funciones
void check_gl_error(){
  int errornum=glGetError();
  while(errornum!=GL_NO_ERROR){
    if (cl_info){
           if(errornum==GL_INVALID_ENUM)
             cout << "GL_INVALID_ENUM" << endl;
      else if(errornum==GL_INVALID_VALUE) 
             cout << "GL_INVALID_VALUE" << endl;
      else if (errornum==GL_INVALID_OPERATION)
             cout << "GL_INVALID_OPERATION" << endl;
      else if (errornum==GL_STACK_OVERFLOW)
             cout << "GL_STACK_OVERFLOW" << endl;
      else if (errornum==GL_STACK_UNDERFLOW)
             cout << "GL_STACK_UNDERFLOW" << endl;
      else if (errornum==GL_OUT_OF_MEMORY)
             cout << "GL_OUT_OF_MEMORY" << endl;
    }
    errornum=glGetError();
  }
}

void salir(){  
  gluDeleteNurbsRenderer(curva);
  delete [] ukv; delete [] pc;
  exit(EXIT_SUCCESS);
}

//--------
// vector de knots
void set_knot_vector(){
  // NURBS de orden uo, con npc puntos de control
  // hay unk=npc+o knots
  // OpenGL usa un knot de mas al ppio y otro al final (por el rango de las funciones de forma)
  int i;
  if (npc<uo) {unk=0; return;}
  unk=npc+uo;
  if (periodico){
    unk+=uo-1;
    for (i=0;i<uo-1;i++) pc[npc+i]=pc[i];
  }
  
  if (BezierBC){
    // condiciones de borde de Bezier 
    // ej de 3er grado
    // 4 iguales a 0 y 4 iguales al final, el primero y el ultimo inutiles
    // los knots van desde 0 a npc-3:
    // 4: 0(000)1111 00(001)111 000(011)11 0000(111)1
    // 5: 0(000)12222 00(001)2222 000(012)222 0000(122)22 00001(222)2
    // 6: 0(000)123333 00(001)23333 000(012)3333 0000(123)333 00001(233)33 000012(333)3
    // orden=grado+1, 
    //    npc-o+2 distintos, que van de 0 a npc-o+1
    //    se repiten o al principio (=0) y o al final (=npc-o+1)
    for (i=0;i<uo-1;i++) {ukv[i]=0; ukv[unk-1-i]=npc-uo+1;} // repetidos
    for (i=0;i<npc-uo+2;i++) ukv[uo-1+i]=i; // distintos (de 0 a npc-o+1)
  }
  else for (i=0;i<unk;i++) ukv[i]=i; // el primero y el ultimo inutiles

  if (cl_info) {
    cout << "knot vector: {" << ukv[0];
    for (i=1;i<unk;i++) cout << "," << ukv[i];
    cout << "}" << endl;
  }
}

void dibuja(){
  if (!npc) return;
  int i;

  // puntos de control
  glColor3fv(pc_c);  
  glBegin(GL_POINTS);
    for (i=0;i<npc;i++) glVertex4fv(pc[i]);
  glEnd();
  if (npc==1) return;
  // poligono de control
  glColor3fv(lc_c); glLineWidth(1);
  if (periodico) glBegin(GL_LINE_LOOP);
  else glBegin(GL_LINE_STRIP);
    for (i=0;i<npc;i++) glVertex4fv(pc[i]);
  glEnd();
  if (npc<uo) return;

  // curva
  glColor3fv(c_c); glLineWidth(2);
  gluBeginCurve(curva);
//  void gluNurbsCurve( GLUnurbsObj *nurbsObj,
//                      GLint nknots, GLfloat *knot, GLint stride,
//                      GLfloat *ctlarray, GLint order, GLenum type );
    gluNurbsCurve(curva,unk,ukv,4,(float*)&pc[0][0],uo,GL_MAP1_VERTEX_4);
  gluEndCurve(curva);
}

//============================================================
// callbacks

//------------------------------------------------------------
// redibuja cada vez que hace un redisplay
void Display_cb() { // Este tiene que estar
  if (mini) return;  

  glClear(GL_COLOR_BUFFER_BIT);  // borra el buffer de pantalla
  dibuja();
  glutSwapBuffers();

  check_gl_error();
}

//------------------------------------------------------------
// Maneja cambios de ancho y alto del screen
void Reshape_cb(int wi, int hi){
  w=wi;h=hi;
//  if (cl_info) cout << "reshape " << w << "x" << h << endl;

  if (!h||!w) {mini=true; return;}
  mini=false;

  glViewport(0,0,w,h); // region donde se dibuja

  // matriz de proyeccion
  glMatrixMode(GL_PROJECTION);  glLoadIdentity();
  glOrtho(0,w,h,0,-1,1);

  Display_cb();
}

//------------------------------------------------------------
// Teclado y Mouse
// GLUT ACTIVE SHIFT Set if the Shift modifier or Caps Lock is active.
// GLUT ACTIVE CTRL  Set if the Ctrl modifier is active.
// GLUT ACTIVE ALT   Set if the Alt modifier is active.

short modifiers=0;  // ctrl, alt, shift (de GLUT)
inline short get_modifiers() {return modifiers=(short)glutGetModifiers();}

// Maneja pulsaciones del teclado (ASCII keys)
// x,y posicion del mouse cuando se teclea
void Keyboard_cb(unsigned char key,int x=0,int y=0) {
  if (key==27){ // escape => exit
    get_modifiers();
    if (!modifiers) salir();
  }
  else if (key=='a'||key=='A'){ // Antialiasing
    antialias=!antialias;
    if (antialias){
      glEnable(GL_POINT_SMOOTH); glEnable(GL_LINE_SMOOTH);
      if (cl_info) cout << "Antialiasing" << endl;
    }
    else {
      glDisable(GL_POINT_SMOOTH); glDisable(GL_LINE_SMOOTH);
      if (cl_info) cout << "Sin Antialiasing" << endl;
    }
  }
  else if (key=='+'){ // lod
    lod++; 
    gluNurbsProperty(curva,GLU_U_STEP,float(lod));
    if (cl_info) {
      cout << "Nivel de Detalle: " << lod  << "               \r"; 
      cout.flush();
    }
  }
  else if (key=='-'){ // lod
    if (lod==2) return;
    lod--; 
    if (cl_info) {
      cout << "Nivel de Detalle: " << lod  << "               \r"; 
      cout.flush();
    }
  }
  else if (key=='b'||key=='B'){ // Bezier
    BezierBC=!BezierBC; if (BezierBC) periodico=false;
    if (cl_info) cout << "Bezier " << ((BezierBC)? "on" : "off") << endl;
    set_knot_vector();
  }
  else if (key=='i'||key=='I'){ // info
    cl_info=!cl_info;
    cout << ((cl_info)? "Info" : "Sin Info") << endl;
    return;
  }
  else if (key=='p'||key=='P'){ // Periodico    
    periodico=!periodico; if (periodico) BezierBC=false;
    if (cl_info) {if (!periodico) cout << "No-"; cout << "Periodico" << endl;}
    set_knot_vector();
  }
  else if (key==127){ // del => borra pt
    if (npc==0) return;
    for (int i=pcsel;i<npc-1;i++)
      pc[i]=pc[i+1];
    npc--;
    if (pcsel==npc) pcsel--;
    if (npc<uo) {      
      uo--;
      if (cl_info) cout << "Grado: " << uo-1 << "               " << endl;
    }
    set_knot_vector();
  }
  Display_cb();
}

// Special keys (non-ASCII)
//  GLUT KEY F[1,12] F[1,12] function key.
//  GLUT KEY LEFT Left directional key.
//  GLUT KEY UP Up directional key.
//  GLUT KEY RIGHT Right directional key.
//  GLUT KEY DOWN Down directional key.
//  GLUT KEY PAGE UP Page up directional key.
//  GLUT KEY PAGE DOWN Page down directional key.
//  GLUT KEY HOME Home directional key.
//  GLUT KEY END End directional key.
//  GLUT KEY INSERT Inset directional key.

// aca es int key
void Special_cb(int key,int xm=0,int ym=0) {
  if (key==GLUT_KEY_F4){ // alt+f4 => exit
    get_modifiers();
    if (modifiers==GLUT_ACTIVE_ALT) salir();
  }
  else if (key==GLUT_KEY_PAGE_UP){ // grado
    if (uo==npc||uo==MAXORDER) return;
    uo++;
    if (cl_info) cout << "Grado: " << uo-1 << "               " << endl;
    set_knot_vector();
  }
  else if (key==GLUT_KEY_PAGE_DOWN){ // grado
    if (uo==1) return;
    uo--;
    if (cl_info) cout << "Grado: " << uo-1 << "               " << endl;
    set_knot_vector();
  }
  Display_cb();
}

// Menu
void Menu_cb(int value)
{
  switch (value){
    case 'a':
      Keyboard_cb('a');
      return;
    case 'b':
      Keyboard_cb('b');
      return;
    case 'p':
      Keyboard_cb('p');
      return;
    case '+':
      Keyboard_cb('+');
      return;
    case '-':
      Keyboard_cb('-');
      return;
    case 127: // del = borra
      Keyboard_cb(127);
      return;
    case (256+GLUT_KEY_PAGE_UP):
      Special_cb(GLUT_KEY_PAGE_UP);
      return;
    case (256+GLUT_KEY_PAGE_DOWN):
      Special_cb(GLUT_KEY_PAGE_DOWN);
      return;
    case 27: // esc = exit
      salir();
  }
}
//------------------------------------------------------------

// Movimientos del mouse
void Motion_cb(int x, int y){ // drag  
  if (modifiers==GLUT_ACTIVE_CTRL){ // cambia el peso
//    float W=W0*pow(10.0,int(20.0*4.0*(yclick-y)/h)/20.0);
    float W=W0+4.0*(yclick-y)/h;
    if (pcsel>=0&&pcsel<npc) pc[pcsel].peso(W);    
    if (cl_info) {
      cout << "Pesos: {";
      for (int k=0;k<npc-1;k++) cout << pc[k][3] << ", ";
      cout << pc[npc-1][3] << "}                                        \r" << flush;
    }
  }
  else{
    // mueve el punto
    // permite drag solo dentro del viewport
    // en caso contrario, sin zoom, no se pueden editar los que esten fuera
    if (x<0) x=0; else if (x>w-1) x=w-1;
    if (y<0) y=0; else if (y>h-1) y=h-1;
    p4f pm(x,y);
    // si se acerco a otro lo pega
    int i;
    for (i=0;i<npc;i++){
      if (i==pcsel) continue;
      if (!pm.cerca2D(pc[i],tol)) continue;
      pm=pc[i];
      break;
    }
    pm.peso(pc[pcsel][3]); pc[pcsel]=pm; // mueve
  }
  if (periodico&&pcsel<uo) pc[npc+pcsel]=pc[pcsel];
  Display_cb();
}

// Clicks del mouse
// GLUT LEFT BUTTON, GLUT MIDDLE BUTTON, or GLUT RIGHT BUTTON
// The state parameter is either GLUT UP or GLUT DOWN
// glutGetModifiers may be called to determine the state of modifier keys
void Mouse_cb(int button, int state, int x, int y){
  if (button==GLUT_LEFT_BUTTON){
    if (state==GLUT_DOWN) {
      p4f pm(x,y);
      int i;
      // verifica tolerancia
      for(i=0;i<npc;i++){
        if (!pm.cerca2D(pc[i],tol)) continue;
        pcsel=i;
        glutMotionFunc(Motion_cb);
        return;
      }
      // no pico cerca de ninguno
      get_modifiers();
      if (modifiers==GLUT_ACTIVE_CTRL){ // cambia el peso
        yclick=y; W0=pc[pcsel][3];
        glutMotionFunc(Motion_cb);
        return;
      }
      if (npc==MAXPC) return;
      // agrega entre medio
      for (i=npc;i>pcsel+1;i--) pc[i]=pc[i-1];
      pc[i]=pm;
      pcsel=i;
      npc++;
      set_knot_vector();
      Display_cb();
      glutMotionFunc(Motion_cb);
      return;
    } // down
    else if (state==GLUT_UP){// fin del drag
      glutMotionFunc(0); // anula el callback para los drags
      modifiers=0;
      return;
    } // up
  } // left
}

//------------------------------------------------------------
int integerv(GLenum pname){
  int value;
  glGetIntegerv(pname,&value);
  return value;
}
#define _PRINT_INT_VALUE(pname) #pname << ": " << integerv(pname) <<endl

//------------------------------------------------------------
// Inicializa GLUT y OpenGL
void initialize() {
	// pide color RGB y double buffering
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE);

  glutInitWindowSize(640,480); glutInitWindowPosition(250,100);

  glutCreateWindow("Curva NURBS"); // crea el main window

  //declara los callbacks
  //los que no se usan no se declaran
  glutDisplayFunc(Display_cb); // redisplays
  glutReshapeFunc(Reshape_cb); // resize
  glutKeyboardFunc(Keyboard_cb); // teclado
  glutSpecialFunc(Special_cb); // ctrl alt y shift
  glutMouseFunc(Mouse_cb); // clicks

  // ========================
  // estado normal del OpenGL
  // ========================
  glClearColor(fondo[0],fondo[1],fondo[2],1);  // color de fondo
  glDrawBuffer(GL_BACK);
  glMatrixMode(GL_MODELVIEW); glLoadIdentity(); // constante
  glShadeModel(GL_FLAT); // no interpola color
  glPointSize(5);

  if (antialias){
    glEnable(GL_POINT_SMOOTH); glEnable(GL_LINE_SMOOTH);
  }
  else {
    glDisable(GL_POINT_SMOOTH); glDisable(GL_LINE_SMOOTH);
  }

  // propiedades de la NURBS
  curva=gluNewNurbsRenderer();
  gluNurbsProperty(curva,GLU_SAMPLING_METHOD,GLU_DOMAIN_DISTANCE); // para usar lod
  gluNurbsProperty(curva,GLU_U_STEP,float(lod));
  gluNurbsProperty(curva,GLU_DISPLAY_MODE,GLU_FILL);

  // maximo orden y arrays de puntos
  glGetIntegerv(GL_MAX_EVAL_ORDER,&MAXORDER);
  pc=new p4f[MAXPC+MAXORDER];
  ukv=new float[MAXPC+2*MAXORDER];
  char menutxt[255];

  glutCreateMenu(Menu_cb);
    glutAddMenuEntry("[b] Condiciones de borde de Bezier ON/OFF", 'b');
    glutAddMenuEntry("[p] Periodica (Cerrada) ON/OFF", 'p');
    sprintf(menutxt,"[PgUp] Aumenta Grado [1,%i]",MAXORDER-1);
    glutAddMenuEntry(menutxt, (256+GLUT_KEY_PAGE_UP));
    sprintf(menutxt,"[PgDn] Disminuye Grado [1,%i]",MAXORDER-1);
    glutAddMenuEntry(menutxt, (256+GLUT_KEY_PAGE_DOWN));
    glutAddMenuEntry("---------------------------", 255);
    glutAddMenuEntry("[Del] Borra Punto de Control", 127);
	  glutAddMenuEntry("[+] +Nivel de Detalle     ", '+');
	  glutAddMenuEntry("[-] -Nivel de Detalle     ", '-');
    glutAddMenuEntry("[a] Antialiasing ON/OFF", 'a');
    glutAddMenuEntry("---------------------------", 255);
    glutAddMenuEntry("[i] Info ON/OFF", 'i');
    glutAddMenuEntry("[Esc] Exit", 27);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  // ========================
  // info
  if (cl_info){
      cout << "\n\n\nPuntos de control:\n"
         << "     Click izq -> Selecciona (tol. 10) o\n"
            "                    Agrega despues del seleccionado\n"
         << "      Drag izq -> Mueve el seleccionado\n"
         << "         <DEL> -> Borra el seleccionado\n"
         << " Ctrl+Drag izq -> Cambia el peso del seleccionado\n"
         << endl;
  }
}

//------------------------------------------------------------
// main
int main(int argc,char** argv) {
  glutInit(&argc,argv);// inicializa glut
  initialize(); // condiciones iniciales de la ventana y OpenGL
  glutMainLoop(); // entra en loop de reconocimiento de eventos
  return 0; //solo para que el builder se quede tranquilo (aqui nunca llega)
}
